#################################
# Shrinkage Averaging Estimaton #
#################################

sae <- function(X,shrink.method=c("Lasso","R.Lasso","Ridge"),averaging=c("OCV","BW_GCV"),
                ycol=1,B=200,Bweights=200,nolambda=100,kfold=5,qv=5,qv2=5,B1=200,
                B2=200,standardize=TRUE, calc.variance=TRUE,print.warnings=TRUE){

shrink.method <- match.arg(shrink.method) # choose methodology
averaging     <- match.arg(averaging)

# User information
if(print.warnings==TRUE & Bweights<25 & averaging=="BW_GCV"){warning("The BW_GCV weights may only be reliable with a higher number of bootstrap replications")}
if(shrink.method=="R.Lasso" & averaging=="BW_GCV"){stop("Sorry. This is not implemented yet! Please swith to `OCV'.")}

# Preparation of datamatrix
if(ycol!=1){X <- cbind(X[,ycol], X[,-ycol]) }
keepnames <- colnames(X)
if(standardize==TRUE){X <-cbind(X[,1],scale(X[,-1]))}
X <- as.data.frame(X)
n <- nrow(X)
m <- ncol(X)
colnames(X)<- paste("x",0:(m-1), sep="")
colnames(X)[1] <- c("y")

# these are our final estimates
shrink_sel     <- NULL       # Shrinkage estimation based on tuning parameter SELECTION
var_shrink_sel <- NULL
shrink_ave     <- NULL       # Shrinkage averaging estimation incorporating tuning parameter SELECTION UNCERTAINTY
var_shrink_ave <- NULL
# ...and this information will also be returned at the end of the function
sae_weights    <-NULL
sel_weights    <-NULL
complexity_parameter <-NULL
variable_importance    <-NULL

################################################################################

# OLS estimate   (as a reference)
mycovariables <- paste("X$x",1:(ncol(X)-1), sep="")
myindependent <- paste("X$y ~")
myformula     <- as.formula(paste(myindependent, paste(mycovariables, collapse="+") ))
ols           <- coefficients(lm(myformula))
var_ols       <- summary(lm(myformula))[[4]][,2]
if(dim(X)[1]<=dim(X)[2]){
ols <- c(rep(0,dim(X)[2]))
var_ols <- c(rep(0,dim(X)[2]))
}

################################################################################
myindex <- function(myvector){
for(i in 1:length(myvector)){
if(min(myvector)==myvector[i]){return(i)}
}
}


#############
### Ridge ###
#############

if(shrink.method=="Ridge"){

# Ridge estimates (for many lambdas)
mycovariables <- paste("X$x",1:(ncol(X)-1), sep="")
myindependent <- paste("X$y ~")
myformula     <- as.formula(paste(myindependent, paste(mycovariables, collapse="+") ))
mylambda      <- c(0,10^(seq(-5, 5, length = nolambda-1)))
myridge       <- MASS::lm.ridge(myformula, data=X, lambda=mylambda)
if(myindex(myridge$GCV)> nolambda-2){print("Chosen no. of lambda may not be suitable! Modify if necessary!")}
if(myindex(myridge$GCV)< 2){print("Chosen no. of lambda may not be suitable! Modify if necessary!")}
ridge_coeff <- coef(myridge)
gcv_inf <- myridge$GCV
complexity_parameter <- mylambda

# Variance of Ridge estimate (for many lambdas, based on the bootstrap)

if(calc.variance==TRUE){
boot.ridge <- function(mydata,indices)
{
mydata <- mydata[indices,]
mycovariables2 <- paste("mydata[,",2:(ncol(mydata)),"]", sep="")
myindependent2 <- paste("mydata[,1] ~")
myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
l<-  c(0,10^(seq(-5, 5, length = nolambda-1)))
myboot <- MASS::lm.ridge(myformula2, data=mydata, lambda=l)
coefficients(myboot)
}
myridge_boot<- boot(as.data.frame(X),boot.ridge,B)
var_ridge <-matrix(apply(myridge_boot$t,2,sd),ncol=ncol(ridge_coeff),nrow=nrow(ridge_coeff))
mean_ridge <-matrix(apply(myridge_boot$t,2,mean),ncol=ncol(ridge_coeff),nrow=nrow(ridge_coeff))
}

        ##################
        # OCV estimators #
        ##################
        if(averaging==c("OCV")){
        
        #  OCV Shrinkage averaging estimation
        
        # Cross Validation
        candidateLambda <- mylambda
        cvRMSE <- rep(0,nolambda)
        cvIndex <- rep(1:kfold,trunc(n/kfold)+1)[1:n]
        cvIndex <- sample(cvIndex)
          for (j in 1:kfold) {
            for (k in (1:nolambda)) {
                myridgedata <- X[cvIndex!=j,]
                myridgedata <- as.data.frame(myridgedata)
                colnames(myridgedata)<- paste("x_cv",0:(m-1), sep="")
                colnames(myridgedata)[1] <- c("y_cv")
                mycovariables2 <- paste("x_cv",1:(ncol(myridgedata)-1), sep="")
                myindependent2 <- paste("y_cv ~")
                myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
                myridge2<-MASS::lm.ridge(myformula2,data=myridgedata,lambda=candidateLambda[k])
                Predata <- X[cvIndex==j,]
                myPred<-   as.matrix(cbind(c(rep(1,nrow(Predata))),Predata[,-1]))%*%matrix(coef(myridge2),ncol=1,nrow=length(coef(myridge2)))
                cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
                }
          }
          cvRMSE <- cvRMSE/kfold
        # Shrinkage with tuning parameter selection based on CV_k 
        weights_minCV <-  c(rep(0, nolambda))
        weights_minCV[myindex(cvRMSE)] <- 1
        
        # Shrinkage averaging estimation = OCV_k
        jma<- function(avdata,mynewlambda,mykfold){

        
        n <- nrow(avdata)
        resmat <- matrix(NA,ncol=nrow(avdata),nrow=mynewlambda)
        mysupernewlambda <- c(0,10^(seq(-5, 5, length = nolambda-1)))

          for(k in 1:mynewlambda){
            for(j in 1:mykfold){

            cvRMSE <- rep(0,mynewlambda)
            cvIndex <- rep(1:mykfold,trunc(n/mykfold)+1)[1:n]
            myridgedata<- avdata[cvIndex!=j,]
            myridgedata <- as.data.frame(myridgedata)
            colnames(myridgedata)<- paste("x_cv",0:(m-1), sep="")
            colnames(myridgedata)[1] <- c("y_cv")
            mycovariables2 <- paste("x_cv",1:(ncol(myridgedata)-1), sep="")
            myindependent2 <- paste("y_cv ~")
            myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
            myridge<- MASS::lm.ridge(myformula2, data=myridgedata, lambda=mysupernewlambda[k])
            Predata <- avdata[cvIndex==j,]
            myPred<-   as.matrix(cbind(c(rep(1,nrow(Predata))),Predata[,-1]))%*%matrix(coef(myridge),ncol=1,nrow=length(coef(myridge)))
            resmat[k,cvIndex==j] <- avdata[cvIndex==j,]$y - myPred
          }}

        
        E <- resmat%*%t(resmat)
        E <- corpcor::make.positive.definite(E)
        Amat <- matrix(c(rep(1,mynewlambda)),nrow=1,ncol=mynewlambda)
        Amat <- t(rbind(Amat,diag(1,mynewlambda)))
        bvec <-c(1,rep(0,mynewlambda))
        dvec <- c(rep(0,mynewlambda))

        weight <- quadprog::solve.QP(E, dvec, Amat, bvec, meq=1)$solution
        weight <- round(weight,digits=4)
        weight <- matrix(c(weight),nrow=1,ncol=mynewlambda)
        colnames(weight)<- paste("w",1:(mynewlambda), sep="")
        rownames(weight)<- paste("OCV weights")
        return(weight)
        }
        
        sae_weights <- suppressWarnings(jma(X,nolambda,kfold))
        sel_weights <- weights_minCV
        
        # Calaculate estimates and their standard error
        shrink_sel  <- weights_minCV%*%ridge_coeff
        shrink_ave  <- sae_weights%*%ridge_coeff
        
        if(calc.variance==TRUE){
          av_1 <- matrix(c(rep(shrink_sel,each=nolambda)),ncol=length(shrink_sel),nrow=nolambda)
          av_3 <- matrix(c(rep(shrink_ave,each=nolambda)),ncol=length(shrink_ave),nrow=nolambda)
          var_shrink_sel  <- (weights_minCV%*%sqrt((ridge_coeff-av_1)^2+var_ridge^2))
          var_shrink_ave   <- (sae_weights%*%sqrt((ridge_coeff-av_3)^2+var_ridge^2))
          }else{
           var_shrink_sel <- rep(NA,length(shrink_sel))
           var_shrink_ave <- rep(NA,length(shrink_ave))
          }
        
        }
        
        #####################
        # BW_GCV estimators #
        #####################
        if(averaging==c("BW_GCV")){
        
        # Shrinkage with tuning parameter selection based on GCV 
        weights_minGCV <-  c(rep(0, nolambda))
        weights_minGCV[myindex(gcv_inf)] <- 1
        
        # Shrinkage averaging based on bootstrapping and GCV
        weights_boot_GCV <- c(rep(0,nolambda))
        boot.GCV.ridge <- function(mydata,indices){
        mydata <- mydata[indices,]
        mycovariables2 <- paste("mydata[,",2:(ncol(mydata)),"]", sep="")
        myindependent2 <- paste("mydata[,1] ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        l <- c(0,10^(seq(-5, 5, length = nolambda-1)))
        myboot <- MASS::lm.ridge(myformula2, data=mydata, lambda=l)
        myboot$GCV
        }
        mygcv_boot<- boot(as.data.frame(X),boot.GCV.ridge,Bweights)
        summary_boot_GCV <- apply(mygcv_boot$t,1,myindex)
        for(i in 1:length(summary_boot_GCV)){
          for(j in 1:length(weights_boot_GCV)){
            if(summary_boot_GCV[i]==j){weights_boot_GCV[j]<-weights_boot_GCV[j]+1}
        }}
        sae_weights <- weights_boot_GCV/Bweights
        sel_weights <- weights_minGCV
        # Calaculate estimates and their standard error
        shrink_sel <-  weights_minGCV%*%ridge_coeff
        shrink_ave  <- sae_weights%*%ridge_coeff
        
        if(calc.variance==TRUE){
          av_1 <- matrix(c(rep(shrink_sel,each=nolambda)),ncol=length(shrink_sel),nrow=nolambda)
          av_3 <- matrix(c(rep(shrink_ave,each=nolambda)),ncol=length(shrink_ave),nrow=nolambda)
          var_shrink_sel  <- (weights_minGCV%*%sqrt((ridge_coeff-av_1)^2+var_ridge^2))
          var_shrink_ave   <- (sae_weights%*%sqrt((ridge_coeff-av_3)^2+var_ridge^2))
          }else{
           var_shrink_sel <- rep(NA,length(shrink_sel))
           var_shrink_ave <- rep(NA,length(shrink_ave))
          }
        
        }

}

################################################################################

#############
### Lasso ###
#############
if(shrink.method=="Lasso")suppressWarnings({  # Warning message related to environment. As far as I understand, this should NOT be a problem


mybound <- (1:nolambda)/(nolambda)

# Lasso estimate
mycovariables <- paste("X$x",1:(ncol(X)-1), sep="")
myindependent <- paste("X$y ~")
myformula     <- as.formula(paste(myindependent, paste(mycovariables, collapse="+") ))
lasso <- l1ce(myformula, X, bound=mybound)
lasso_coeff <- coefficients(lasso)
lasso_res <- residuals(lasso)
gcv_inf <- gcv(lasso, type="Tibshirani")
complexity_parameter <- mybound


# Variance of lasso estimate
if(calc.variance==TRUE){
boot.lasso <- function(mydata,indices)
{
mydata <- mydata[indices,]
mycovariables2 <- paste("mydata[,",2:(ncol(mydata)),"]", sep="")
myindependent2 <- paste("mydata[,1] ~")
myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
myboot <- l1ce(myformula2, data=mydata, bound=mybound)
coefficients(myboot)
}
mylasso_boot<- boot(as.data.frame(X),boot.lasso,B)
var_lasso <-matrix(apply(mylasso_boot$t,2,sd),ncol=ncol(lasso_coeff),nrow=nrow(lasso_coeff))
mean_lasso <-matrix(apply(mylasso_boot$t,2,mean),ncol=ncol(lasso_coeff),nrow=nrow(lasso_coeff))
}

                  
                  
        ##################
        # OCV estimators #
        ##################
        if(averaging==c("OCV")){
        
        #  OCV Shrinkage averaging estimation
        
        ## Cross Validation
        candidateBounds <- gcv_inf[,1]
        cvRMSE <- rep(0,nolambda)

        cvIndex <- rep(1:kfold,trunc(n/kfold)+1)[1:n]
        cvIndex <- sample(cvIndex)
          for (j in 1:kfold) {
            for (k in (1:nolambda)) {
        mylassodata<- X[cvIndex!=j,]
        mylassodata <- as.data.frame(mylassodata)
        colnames(mylassodata)<- paste("x_cv",0:(m-1), sep="")
        colnames(mylassodata)[1] <- c("y_cv")
        mycovariables2 <- paste("x_cv",1:(ncol(mylassodata)-1), sep="")
        myindependent2 <- paste("y_cv ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        mylasso<-l1ce(myformula2,data=mylassodata,bound=candidateBounds[k])
        Predata <- X[cvIndex==j,]
        colnames(Predata) <- colnames(mylassodata)
        myPred<-predict(mylasso,newdata=Predata)
        cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
          }}

        cvRMSE <- cvRMSE/kfold
        # Shrinkage with tuning parameter selection based on CV_k 
        weights_minCV <-  c(rep(0, nolambda))
        weights_minCV[myindex(cvRMSE)] <- 1
        
        # Shrinkage averaging estimation = OCV_k
        jma<- function(avdata,mynewlambda,mykfold){

        n <- nrow(avdata)
        resmat <- matrix(NA,ncol=nrow(avdata),nrow=mynewlambda)
        mysupernewbound <- (1:nolambda)/(nolambda)

          for(k in 1:mynewlambda){
            for(j in 1:mykfold){

        cvRMSE <- rep(0,mynewlambda)
        cvIndex <- rep(1:mykfold,trunc(n/mykfold)+1)[1:n]
        mylassodata<- X[cvIndex!=j,]
        mylassodata <- as.data.frame(mylassodata)
        colnames(mylassodata)<- paste("x_cv",0:(m-1), sep="")
        colnames(mylassodata)[1] <- c("y_cv")
        mycovariables2 <- paste("x_cv",1:(ncol(mylassodata)-1), sep="")
        myindependent2 <- paste("y_cv ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        mylasso<-l1ce(myformula2,data=mylassodata,bound=mysupernewbound[k])
        Predata <- X[cvIndex==j,]
        colnames(Predata) <- colnames(mylassodata)
        myPred<-predict(mylasso,newdata=Predata)
        cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
        resmat[k,cvIndex==j] <- avdata[cvIndex==j,]$y - myPred
        }}

        E <- resmat%*%t(resmat)
        E <- make.positive.definite(E) # is normally not needed...just in case there is a problem
        Amat <- matrix(c(rep(1,mynewlambda)),nrow=1,ncol=mynewlambda)
        Amat <- t(rbind(Amat,diag(1,mynewlambda)))
        bvec <-c(1,rep(0,mynewlambda))
        dvec <- c(rep(0,mynewlambda))

        weight <- solve.QP(E, dvec, Amat, bvec, meq=1)$solution
        weight <- round(weight,digits=4)
        weight <- matrix(c(weight),nrow=1,ncol=mynewlambda)
        colnames(weight)<- paste("w",1:(mynewlambda), sep="")
        rownames(weight)<- paste("OCV weights")
        return(weight)
        }

        sae_weights <- jma(X,nolambda,kfold)
        sel_weights <- weights_minCV
        # Calaculate estimates and their standard error
        shrink_sel  <- weights_minCV%*%lasso_coeff
        shrink_ave  <- sae_weights%*%lasso_coeff
        
        if(calc.variance==TRUE){
          av_1 <- matrix(c(rep(shrink_sel,each=nolambda)),ncol=length(shrink_sel),nrow=nolambda)
          av_3 <- matrix(c(rep(shrink_ave,each=nolambda)),ncol=length(shrink_ave),nrow=nolambda)
          var_shrink_sel  <- (weights_minCV%*%sqrt((lasso_coeff-av_1)^2+var_lasso^2))
          var_shrink_ave   <- (sae_weights%*%sqrt((lasso_coeff-av_3)^2+var_lasso^2))
          }else{
           var_shrink_sel <- rep(NA,length(shrink_sel))
           var_shrink_ave <- rep(NA,length(shrink_ave))
          }
          
        # Calculate Variable Importance Measure
        index_imp <- matrix(rep(0,ncol(lasso_coeff)*nrow(lasso_coeff)),ncol=ncol(lasso_coeff),nrow=nrow(lasso_coeff))
        for(j in 1:ncol(lasso_coeff)){
        for(i in 1:nrow(lasso_coeff)){
        if(abs(lasso_coeff[i,j])>1e-05){index_imp[i,j]<-1}
        }}
        variable_importance<- t(index_imp)%*%matrix(sae_weights,ncol=1, nrow=length(sae_weights))
        rownames(variable_importance) <- keepnames
        colnames(variable_importance) <- "Importance"
        
        }
        
        #####################
        # BW_GCV estimators #
        #####################
        if(averaging==c("BW_GCV")){
        
        # Shrinkage with tuning parameter selection based on GCV 
        weights_minGCV <-  c(rep(0, nolambda))
        weights_minGCV[myindex(gcv_inf[,4])] <- 1
        
        # Shrinkage averaging based on bootstrapping and GCV
        weights_boot_GCV_gen <- c(rep(0,nolambda))
        boot.GCV <- function(mydata,indices)
        {
        mydata <- mydata[indices,]
        mycovariables2 <- paste("mydata[,",2:(ncol(mydata)),"]", sep="")
        myindependent2 <- paste("mydata[,1] ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        myboot <- l1ce(myformula2, data=mydata, bound=mybound)
        (gcv(myboot,type="Tibshirani"))[,4]
        }
        mygcv_boot<- boot(as.data.frame(X),boot.GCV,Bweights)
        summary_boot_GCV <- apply(mygcv_boot$t,1,myindex)
          for(i in 1:length(summary_boot_GCV)){
             for(j in 1:length(weights_boot_GCV_gen)){
                if(summary_boot_GCV[i]==j){weights_boot_GCV_gen[j]<-weights_boot_GCV_gen[j]+1}
          }}
        sae_weights  <- weights_boot_GCV_gen/Bweights
        sel_weights <- weights_minGCV
      
        # Calaculate estimates and their standard error
        shrink_sel <-  weights_minGCV%*%lasso_coeff
        shrink_ave  <- sae_weights%*%lasso_coeff
        
        if(calc.variance==TRUE){
          av_1 <- matrix(c(rep(shrink_sel,each=nolambda)),ncol=length(shrink_sel),nrow=nolambda)
          av_3 <- matrix(c(rep(shrink_ave,each=nolambda)),ncol=length(shrink_ave),nrow=nolambda)
          var_shrink_sel  <- (weights_minGCV%*%sqrt((lasso_coeff-av_1)^2+var_lasso^2))
          var_shrink_ave   <- (sae_weights%*%sqrt((lasso_coeff-av_3)^2+var_lasso^2))
          }else{
           var_shrink_sel <- rep(NA,length(shrink_sel))
           var_shrink_ave <- rep(NA,length(shrink_ave))
          }
        
        # Calculate Variable Importance Measure
        index_imp <- matrix(rep(0,ncol(lasso_coeff)*nrow(lasso_coeff)),ncol=ncol(lasso_coeff),nrow=nrow(lasso_coeff))
        for(j in 1:ncol(lasso_coeff)){
        for(i in 1:nrow(lasso_coeff)){
        if(abs(lasso_coeff[i,j])>1e-05){index_imp[i,j]<-1}
        }}
        variable_importance<- t(index_imp)%*%matrix(sae_weights,ncol=1, nrow=length(sae_weights))
        rownames(variable_importance) <- keepnames
        colnames(variable_importance) <- "Importance"
        
        }
 })
################################################################################


if(shrink.method=="R.Lasso")suppressWarnings({
 
 
# Step 1 Importance measure
boot.importance <- function(mydata,indices)
{
mydata <- mydata[indices,]
mybound <- (1:nolambda)/(nolambda)
randomvariables <- sample(c(2:m),qv)
mycovariables <- paste("mydata[,",randomvariables,"]", sep="")
myindependent <- paste("mydata[,1] ~")
myformula     <- as.formula(paste(myindependent, paste(mycovariables, collapse="+") ))
myboot <- l1ce(myformula, data=mydata, bound=mybound)
  # Cross Validation
  candidateBounds <- gcv(myboot)[,1]
  cvRMSE <- rep(0,nolambda)
  cvIndex <- rep(1:kfold,trunc(n/kfold)+1)[1:n]
  cvIndex <- sample(cvIndex)
    for (j in 1:kfold) {
      for (k in (1:nolambda)) {
        mylassodata<- mydata[cvIndex!=j,]
        mylassodata <- as.data.frame(mylassodata)
        colnames(mylassodata)<- paste("x_cv",0:(m-1), sep="")
        colnames(mylassodata)[1] <- c("y_cv")
        mycovariables2 <- paste("x_cv",randomvariables-1, sep="")
        myindependent2 <- paste("y_cv ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        mylasso<-l1ce(myformula2,data=mylassodata,bound=candidateBounds[k])
        Predata <- mydata[cvIndex==j,]
        colnames(Predata) <- colnames(mylassodata)
        myPred<-predict(mylasso,newdata=Predata)
        cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
      }
  }
  cvRMSE <- cvRMSE/kfold
  #
mybootcoefficients <- coefficients(myboot)[myindex(cvRMSE),]
rlassocoeff <- c(rep(0,m))
rlassocoeff[c(1,randomvariables)] <- mybootcoefficients
rlassocoeff
}
myimportance_boot<- boot(as.data.frame(X),boot.importance,B1)
importance <- abs(apply(myimportance_boot$t,2,mean))
selectionprob <- importance[-1]/sum(importance[-1])

# Step 2; Estimate
boot.select <- function(mydata,indices)
{
mydata <- mydata[indices,]
mybound <- (1:nolambda)/(nolambda)
randomvariables2 <- sample(c(2:m),qv2,prob=selectionprob+1e-05)
mycovariables2 <- paste("mydata[,",randomvariables2,"]", sep="")
myindependent2 <- paste("mydata[,1] ~")
myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
myboot2 <- l1ce(myformula2, data=mydata, bound=mybound)
# My nice OCV methodology
  jma<- function(avdata,mynewlambda,mykfold){
  n <- nrow(avdata)
  resmat <- matrix(NA,ncol=nrow(avdata),nrow=mynewlambda)
  mysupernewbound <- (1:nolambda)/(nolambda)
  for(k in 1:mynewlambda){
    for(j in 1:mykfold){
        cvRMSE <- rep(0,mynewlambda)
        cvIndex <- rep(1:mykfold,trunc(n/mykfold)+1)[1:n]
        mylassodata<- avdata[cvIndex!=j,]
        mylassodata <- as.data.frame(mylassodata)
        colnames(mylassodata)<- paste("x_cv",0:(m-1), sep="")
        colnames(mylassodata)[1] <- c("y_cv")
        mycovariables2 <- paste("x_cv",randomvariables2-1, sep="")
        myindependent2 <- paste("y_cv ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        mylasso<-l1ce(myformula2,data=mylassodata,bound=mysupernewbound[k])
        Predata <- avdata[cvIndex==j,]
        colnames(Predata) <- colnames(mylassodata)
        myPred<-predict(mylasso,newdata=Predata)
        cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
        resmat[k,cvIndex==j] <- avdata[cvIndex==j,]$y - myPred
  }}
  E <- resmat%*%t(resmat)
  E <- make.positive.definite(E)
  Amat <- matrix(c(rep(1,mynewlambda)),nrow=1,ncol=mynewlambda)
  Amat <- t(rbind(Amat,diag(1,mynewlambda)))
  bvec <-c(1,rep(0,mynewlambda))
  dvec <- c(rep(0,mynewlambda))
  weight <- solve.QP(E, dvec, Amat, bvec, meq=1)$solution
  weight <- round(weight,digits=4)
  weight <- matrix(c(weight),nrow=1,ncol=mynewlambda)
  colnames(weight)<- paste("w",1:(mynewlambda), sep="")
  rownames(weight)<- paste("Jackknife weights")
  return(weight)
  }
  weights_JN <- jma(mydata,nolambda,kfold)
  # Cross Validation
  candidateBounds <- gcv(myboot2)[,1]
  cvRMSE <- rep(0,nolambda)
  cvIndex <- rep(1:kfold,trunc(n/kfold)+1)[1:n]
  cvIndex <- sample(cvIndex)
    for (j in 1:kfold) {
      for (k in (1:nolambda)) {
        mylassodata<- mydata[cvIndex!=j,]
        mylassodata <- as.data.frame(mylassodata)
        colnames(mylassodata)<- paste("x_cv",0:(m-1), sep="")
        colnames(mylassodata)[1] <- c("y_cv")
        mycovariables2 <- paste("x_cv",randomvariables2-1, sep="")
        myindependent2 <- paste("y_cv ~")
        myformula2     <- as.formula(paste(myindependent2, paste(mycovariables2, collapse="+") ))
        mylasso<-l1ce(myformula2,data=mylassodata,bound=candidateBounds[k])
        Predata <- mydata[cvIndex==j,]
        colnames(Predata) <- colnames(mylassodata)
        myPred<-predict(mylasso,newdata=Predata)
        cvRMSE[k] <- cvRMSE[k] + mean((X[cvIndex==j,]$y - myPred)^2)
      }
  }
  cvRMSE <- cvRMSE/kfold
  #
mybootcoefficients <- coefficients(myboot2)[myindex(cvRMSE),]
myOCVcoefficients <-   weights_JN%*%coefficients(myboot2)
rlassocoeff <- c(rep(0,m))
rlassocoeff[c(1,randomvariables2)] <- mybootcoefficients
rlassocoeff_OCV <-  c(rep(0,m))
rlassocoeff_OCV[c(1,randomvariables2)] <- myOCVcoefficients
c(rlassocoeff,rlassocoeff_OCV)
}
myselect_boot<- boot(as.data.frame(X),boot.select,B2)

rlasso <- apply(myselect_boot$t,2,mean)
rlasso_OCV <- rlasso[((length(rlasso)/2)+1):length(rlasso)]
rlasso <- rlasso[1:(length(rlasso)/2)]
names(rlasso) <- keepnames
names(rlasso)[1] <- c("Interc.")
names(rlasso_OCV) <- keepnames
names(rlasso_OCV)[1] <- c("Interc.")

shrink_ave <- rlasso_OCV
var_shrink_ave <- rep(NA,length(shrink_ave))
shrink_sel <- rlasso
var_shrink_sel <- rep(NA,length(shrink_sel))
sae_weights <- "For Random Lasso there are B2 different OCV weights (for each bootstrap sample). They are not reported."
sel_weights <- "For Random Lasso many different models were chosen (for each of the B1 and B2 boostrap samples). These results are not reported)" 
complexity_parameter <- (1:nolambda)/(nolambda)
})
 
 
################################################################################

# Final results
if(print.warnings==TRUE){cat(paste("You performed",shrink.method,"averaging based on",averaging,"\n"))}
mycoefficients <- matrix(c(shrink_ave,var_shrink_ave,shrink_sel,var_shrink_sel,ols,var_ols),nrow=m,ncol=6)
rownames(mycoefficients)<-c("Interc.",keepnames[-1])
colnames(mycoefficients)<- c("SAE est","se",shrink.method,"se","OLS","se")
if(shrink.method=="Lasso"){rownames(variable_importance)[1]<-"Interc."}

################################################################################

# Return our results

res= list(coefficients=mycoefficients,
          variable.importance = variable_importance,
          sae.weights = sae_weights,
          sel.weights = sel_weights,
          complexity.parameter = complexity_parameter, 
          used.method = shrink.method         
)

class(res) <- "sae"
res


}